"""Used to initialise the project, with the correct scaffolding."""

import os
import subprocess
import sys
from pathlib import Path

import click

from ..errors import STBError
from ..ui import log

_TEMPLATE_URL = "https://github.com/pradyunsg/sphinx-theme-template/archive/refs/heads/sphinx-theme-template-stable.zip"


class NewCommand:
    """Create a new project."""

    interface: list[click.Parameter] = [
        click.Argument(
            ["source_directory"],
            required=True,
            type=click.Path(
                exists=False,
                file_okay=False,
                dir_okay=True,
                writable=True,
                path_type=Path,
            ),
        ),
    ]

    def run(self, source_directory: Path) -> int:
        log(f"[yellow]#[/] Looking at [magenta]{source_directory}[/]")

        pyproject_toml = source_directory / "pyproject.toml"
        setup_py = source_directory / "setup.py"
        if pyproject_toml.exists() or setup_py.exists():
            raise STBError(
                code="can-not-overwrite-existing-python-project",
                message="Refusing to generate a new project in provided directory.",
                causes=[
                    "This directory contains a Python project, which will not be "
                    "overwritten."
                ],
                hint_stmt=(
                    "This command should be used on empty/non-existing directories."
                ),
            )

        command = [
            sys.executable,
            "-m",
            "cookiecutter",
            "-o",
            os.fsdecode(source_directory),
            _TEMPLATE_URL,
        ]
        log(f"[magenta]$[/] python {' '.join(command[1:])}")
        try:
            subprocess.run(command, check=True)
        except subprocess.CalledProcessError as error:
            raise STBError(
                code="cookiecutter-failed",
                message="Cookiecutter failed to generate the project.",
                causes=[f"Exit code: {error.returncode}"],
                note_stmt="cookiecutter's failure output is available above.",
                hint_stmt=(
                    "[b]@pradyunsg[/] might still need to write/update the supporting "
                    "cookiecutter template. :sweat_smile:"
                ),
            ) from error

        return 0



import subprocess
import sys
from pathlib import Path
from unittest import mock

from click import Group
from click.testing import CliRunner

from sphinx_theme_builder._internal.cli.new import _TEMPLATE_URL
from sphinx_theme_builder._internal.errors import STBError


class TestNewCommand:
    """`stb new`"""

    def test_aborts_when_setup_py_exists(self, runner: CliRunner, cli: Group) -> None:
        with runner.isolated_filesystem() as directory:
            (Path(directory) / "setup.py").write_text("")

            with mock.patch("subprocess.run") as mocked_run:
                process = runner.invoke(cli, ["new", directory])

        assert mocked_run.call_count == 0
        assert process.exit_code == 1, process

    def test_aborts_when_pyproject_toml_exists(
        self, runner: CliRunner, cli: Group
    ) -> None:
        with runner.isolated_filesystem() as directory:
            (Path(directory) / "pyproject.toml").write_text("")

            with mock.patch("subprocess.run") as mocked_run:
                process = runner.invoke(cli, ["new", directory])

        assert mocked_run.call_count == 0
        assert process.exit_code == 1, process

    def test_calls_cookiecutter(self, runner: CliRunner, cli: Group) -> None:
        with runner.isolated_filesystem() as directory:
            with mock.patch("subprocess.run") as mocked_run:
                process = runner.invoke(cli, ["new", directory])

        assert process.exit_code == 0, process
        assert mocked_run.call_count == 1
        assert mocked_run.call_args == mock.call(
            [sys.executable, "-m", "cookiecutter", "-o", directory, _TEMPLATE_URL],
            check=True,
        )

    def test_cookiecutter_failure(self, runner: CliRunner, cli: Group) -> None:
        with runner.isolated_filesystem() as directory:
            with mock.patch("subprocess.run") as mocked_run:
                mocked_run.side_effect = subprocess.CalledProcessError(
                    returncode=2, cmd="placeholder"
                )
                process = runner.invoke(cli, ["new", directory])

        assert mocked_run.call_count == 1
        assert mocked_run.call_args == mock.call(
            [sys.executable, "-m", "cookiecutter", "-o", directory, _TEMPLATE_URL],
            check=True,
        )

        assert process.exit_code == 1, process
        assert isinstance(process.exception, STBError)



import sys
import textwrap
from pathlib import Path
from typing import Any
from unittest import mock

import pytest
from packaging.version import Version

from sphinx_theme_builder._internal.project import (
    ImproperProjectMetadata,
    InvalidProjectStructure,
    Project,
    read_toml_file,
)


@mock.patch(f"{'tomllib' if sys.version_info > (3, 11) else 'tomli'}.load")
def test_read_toml_file(patched_load: mock.Mock, tmp_path: Path) -> None:
    # GIVEN
    file = tmp_path / "foo.toml"
    file.write_text("")

    # WHEN
    read_toml_file(file)

    # THEN
    patched_load.assert_called_once()
    assert patched_load.call_args[0][0].name == str(file)


class TestProjectFromPath:
    def test_works_on_valid_pyproject(self, tmp_path: Path) -> None:
        # GIVEN
        (tmp_path / "pyproject.toml").write_text(
            textwrap.dedent(
                """
                [project]
                name = 'magic-project'
                version = '1.0'
                license = { text = "MIT" }

                [tool.sphinx-theme-builder]
                node-version = "16.13.0"
                """
            )
        )
        (tmp_path / "src" / "magic_project").mkdir(parents=True)
        (tmp_path / "src" / "magic_project" / "__init__.py").write_text("")

        # WHEN
        project = Project.from_path(tmp_path)

        # THEN
        assert project.snake_name == "magic_project"
        assert project.kebab_name == "magic-project"
        assert project.location == tmp_path

        assert project.metadata
        assert project.metadata.version == Version("1.0")

        assert project.python_package_path == tmp_path / "src" / "magic_project"
        assert project.theme_path == (
            tmp_path / "src" / "magic_project" / "theme" / "magic-project"
        )
        assert project.theme_conf_path == (
            tmp_path
            / "src"
            / "magic_project"
            / "theme"
            / "magic-project"
            / "theme.conf"
        )
        assert project.theme_static_path == (
            tmp_path / "src" / "magic_project" / "theme" / "magic-project" / "static"
        )
        assert project.output_script_path == (
            project.theme_static_path / "scripts" / "magic-project.js"
        )
        assert project.output_stylesheet_path == (
            project.theme_static_path / "styles" / "magic-project.css"
        )

        assert project.assets_path == tmp_path / "src" / "magic_project" / "assets"
        assert project.input_stylesheets_path == (
            tmp_path / "src" / "magic_project" / "assets" / "styles"
        )
        assert project.input_scripts_path == (
            tmp_path / "src" / "magic_project" / "assets" / "scripts"
        )

    def test_rejects_without_pyproject_toml(self, tmp_path: Path) -> None:
        # GIVEN
        # an empty directory

        # WHEN
        with pytest.raises(InvalidProjectStructure) as ctx:
            Project.from_path(tmp_path)

        # THEN
        error = ctx.value
        assert "Could not find a `pyproject.toml`" in error.message
        [cause] = error.causes
        assert str(tmp_path) in cause
        assert error.code == "pyproject-missing"

    def test_rejects_on_invalid_pyproject_toml(self, tmp_path: Path) -> None:
        # GIVEN
        (tmp_path / "pyproject.toml").write_text("key = value")

        # WHEN
        with pytest.raises(ImproperProjectMetadata) as ctx:
            Project.from_path(tmp_path)

        # THEN
        error = ctx.value
        assert "Could not parse `pyproject.toml`" in error.message
        [cause] = error.causes
        assert str(tmp_path) in cause
        assert "pyproject.toml" in cause
        assert "line 1, column 7" in cause
        assert error.code == "pyproject-could-not-parse"

    def test_rejects_without_project_table(self, tmp_path: Path) -> None:
        # GIVEN
        (tmp_path / "pyproject.toml").write_text("key = 1")

        # WHEN
        with pytest.raises(ImproperProjectMetadata) as ctx:
            Project.from_path(tmp_path)

        # THEN
        error = ctx.value
        assert "Could not find [project] table" in error.message
        [cause] = error.causes
        assert str(tmp_path) in cause
        assert "pyproject.toml" in cause
        assert error.code == "pyproject-no-project-table"

    def test_rejects_without_name(self, tmp_path: Path) -> None:
        # GIVEN
        (tmp_path / "pyproject.toml").write_text("[project]\nkey = 1")

        # WHEN
        with pytest.raises(ImproperProjectMetadata) as ctx:
            Project.from_path(tmp_path)

        # THEN
        error = ctx.value
        assert "Could not find `name`" in error.message
        assert "[project] table" in error.message
        [cause] = error.causes
        assert "pyproject.toml" in cause
        assert error.code == "pyproject-no-name-in-project-table"

    def test_rejects_non_canonical_names(self, tmp_path: Path) -> None:
        # GIVEN
        (tmp_path / "pyproject.toml").write_text("[project]\nname = 'MAGIC'")

        # WHEN
        with pytest.raises(ImproperProjectMetadata) as ctx:
            Project.from_path(tmp_path)

        # THEN
        error = ctx.value
        assert (
            "Found non-canonical `name` declared in the [project] table"
            in error.message
        )
        [cause] = error.causes
        assert "pyproject.toml" in cause
        assert error.code == "pyproject-non-canonical-name"

    def test_works_with_proper_static_version(self, tmp_path: Path) -> None:
        # GIVEN
        (tmp_path / "pyproject.toml").write_text(
            textwrap.dedent(
                """
                [project]
                name = "magic"
                version = "0.1.2"
                license = { text = "MIT" }

                [tool.sphinx-theme-builder]
                node-version = "16.13.0"
                """
            )
        )
        (tmp_path / "src" / "magic").mkdir(parents=True)
        (tmp_path / "src" / "magic" / "__init__.py").write_text("")

        # WHEN
        project = Project.from_path(tmp_path)

        # THEN
        assert project.metadata.version == Version("0.1.2")

    def test_works_with_proper_dynamic_version(self, tmp_path: Path) -> None:
        # GIVEN
        (tmp_path / "pyproject.toml").write_text(
            textwrap.dedent(
                """
                [project]
                name = "magic"
                dynamic = ["version"]
                license = { text = "MIT" }

                [tool.sphinx-theme-builder]
                node-version = "16.13.0"
                """
            )
        )
        (tmp_path / "src" / "magic").mkdir(parents=True)
        (tmp_path / "src" / "magic" / "__init__.py").write_text(
            textwrap.dedent(
                """
                version = "2.3.4"  # not really, we just ignore this.
                __version__ = "1.2.3"
                """
            )
        )

        # WHEN
        project = Project.from_path(tmp_path)

        # THEN
        assert project.metadata.version == Version("1.2.3")

    def test_rejects_when_no_version_is_declared(self, tmp_path: Path) -> None:
        # GIVEN
        (tmp_path / "pyproject.toml").write_text(
            textwrap.dedent(
                """
                [project]
                name = "magic"
                """
            )
        )
        (tmp_path / "src" / "magic").mkdir(parents=True)
        (tmp_path / "src" / "magic" / "__init__.py").write_text("")

        # WHEN
        with pytest.raises(InvalidProjectStructure) as ctx:
            Project.from_path(tmp_path)

        # THEN
        error = ctx.value
        assert "No version declaration found for project." in error.message
        [cause] = error.causes
        assert "__init__.py" in cause
        assert "pyproject.toml" in cause
        assert error.hint_stmt
        assert "forget" in error.hint_stmt
        assert "?" in error.hint_stmt
        assert error.code == "project-no-version-declaration"

    def test_rejects_when_no_python_file_is_available(self, tmp_path: Path) -> None:
        # GIVEN
        (tmp_path / "pyproject.toml").write_text(
            textwrap.dedent(
                """
                [project]
                name = "magic"
                version = "1.0.0"
                """
            )
        )

        # WHEN
        with pytest.raises(InvalidProjectStructure) as ctx:
            Project.from_path(tmp_path)

        # THEN
        error = ctx.value
        assert "__init__.py is missing" in error.message
        [cause] = error.causes
        assert "project name is magic" in cause
        assert error.code == "project-init-missing"

    def test_rejects_with_double_declaration(self, tmp_path: Path) -> None:
        # GIVEN
        (tmp_path / "pyproject.toml").write_text(
            textwrap.dedent(
                """
                [project]
                name = "magic"
                version = "1.2.3"
                """
            )
        )
        (tmp_path / "src" / "magic").mkdir(parents=True)
        (tmp_path / "src" / "magic" / "__init__.py").write_text('__version__ = "2.3.4"')

        # WHEN
        with pytest.raises(InvalidProjectStructure) as ctx:
            Project.from_path(tmp_path)

        # THEN
        error = ctx.value
        assert "Found version declaration in both" in error.message
        assert "pyproject.toml" in error.message
        assert "__init__.py" in error.message
        assert error.code == "project-double-version-declaration"

    def test_rejects_dynamic_with_version_in_pyproject(self, tmp_path: Path) -> None:
        # GIVEN
        (tmp_path / "pyproject.toml").write_text(
            textwrap.dedent(
                """
                [project]
                name = "magic"
                version = "2.3.4"
                dynamic = ["version"]
                """
            )
        )
        (tmp_path / "src" / "magic").mkdir(parents=True)
        (tmp_path / "src" / "magic" / "__init__.py").write_text("")

        # WHEN
        with pytest.raises(ImproperProjectMetadata) as ctx:
            Project.from_path(tmp_path)

        # THEN
        error = ctx.value
        assert "dynamic" in error.message
        assert "version" in error.message
        assert "pyproject.toml" in error.message
        [cause] = error.causes
        assert "2.3.4" in cause
        assert error.hint_stmt
        assert "removing `version`" in error.hint_stmt
        assert error.code == "project-improper-dynamic-version"

    def test_rejects_no_dynamic_with_version_in_python_file(
        self, tmp_path: Path
    ) -> None:
        # GIVEN
        (tmp_path / "pyproject.toml").write_text(
            textwrap.dedent(
                """
                [project]
                name = "magic"
                """
            )
        )
        (tmp_path / "src" / "magic").mkdir(parents=True)
        (tmp_path / "src" / "magic" / "__init__.py").write_text('__version__ = "1.2.3"')

        # WHEN
        with pytest.raises(ImproperProjectMetadata) as ctx:
            Project.from_path(tmp_path)

        # THEN
        error = ctx.value
        assert "Found version in `__init__.py`" in error.message
        assert '"version" in `project.dynamic' in error.message
        [cause] = error.causes
        assert "1.2.3" in cause
        assert error.hint_stmt
        assert error.code == "project-missing-dynamic-version"

    @pytest.mark.parametrize("value", [1, 1.0, []])
    def test_rejects_non_string_static_versions_toml(
        self, tmp_path: Path, value: Any
    ) -> None:
        # GIVEN
        (tmp_path / "pyproject.toml").write_text(
            textwrap.dedent(
                f"""
                [project]
                name = "magic"
                version = {value}
                """
            )
        )
        (tmp_path / "src" / "magic").mkdir(parents=True)
        (tmp_path / "src" / "magic" / "__init__.py").write_text("")

        # WHEN
        with pytest.raises(ImproperProjectMetadata) as ctx:
            Project.from_path(tmp_path)

        # THEN
        error = ctx.value
        assert "Expected " in error.message
        assert "a string" in error.message
        [cause] = error.causes
        assert f"Got {value}" in cause
        assert f"{type(value)}" in cause
        assert error.code == "pyproject-invalid-version"

    def test_rejects_invalid_versions(self, tmp_path: Path) -> None:
        # GIVEN
        (tmp_path / "pyproject.toml").write_text(
            textwrap.dedent(
                """
                [project]
                name = "magic"
                version = "asdf"
                """
            )
        )
        (tmp_path / "src" / "magic").mkdir(parents=True)
        (tmp_path / "src" / "magic" / "__init__.py").write_text("")

        # WHEN
        with pytest.raises(ImproperProjectMetadata) as ctx:
            Project.from_path(tmp_path)

        # THEN
        error = ctx.value
        assert "invalid version" in error.message
        assert "asdf" in error.message
        [cause] = error.causes
        assert "pyproject.toml" in cause
        assert error.code == "project-invalid-version"

    def test_rejects_invalid_python_files(self, tmp_path: Path) -> None:
        # GIVEN
        (tmp_path / "pyproject.toml").write_text(
            textwrap.dedent(
                """
                [project]
                name = "magic"
                dynamic = ["version"]
                """
            )
        )
        (tmp_path / "src" / "magic").mkdir(parents=True)
        (tmp_path / "src" / "magic" / "__init__.py").write_text(")")

        # WHEN
        with pytest.raises(InvalidProjectStructure) as ctx:
            Project.from_path(tmp_path)

        # THEN
        error = ctx.value
        assert "Could not parse" in error.message
        [cause] = error.causes
        assert "__init__.py" in cause
        assert "SyntaxError" in cause
        assert error.code == "project-init-invalid-syntax"

    def test_rejects_when_no_license_declared(self, tmp_path: Path) -> None:
        # GIVEN
        (tmp_path / "pyproject.toml").write_text(
            textwrap.dedent(
                """
                [project]
                name = "magic"
                version = "1.0.0"

                [tool.sphinx-theme-builder]
                node-version = "16.13.0"
                """
            )
        )
        (tmp_path / "src" / "magic").mkdir(parents=True)
        (tmp_path / "src" / "magic" / "__init__.py").write_text("")

        # WHEN
        with pytest.raises(ImproperProjectMetadata) as ctx:
            Project.from_path(tmp_path)

        # THEN
        error = ctx.value
        assert "No license is declared" in error.message
        assert error.code == "no-license-declared"

    def test_rejects_missing_license_file(self, tmp_path: Path) -> None:
        # GIVEN
        (tmp_path / "pyproject.toml").write_text(
            textwrap.dedent(
                """
                [project]
                name = "magic"
                version = "1.0.0"
                license = { file = "LICENSE" }

                [tool.sphinx-theme-builder]
                node-version = "16.13.0"
                """
            )
        )
        (tmp_path / "src" / "magic").mkdir(parents=True)
        (tmp_path / "src" / "magic" / "__init__.py").write_text("")

        # WHEN
        with pytest.raises(InvalidProjectStructure) as ctx:
            Project.from_path(tmp_path)

        # THEN
        error = ctx.value
        assert "Provided project metadata is not valid." in error.message
        [cause] = error.causes
        assert "License file not found" in cause
        assert "LICENSE" in cause
        assert error.code == "invalid-pyproject-toml"

    def test_works_with_legacy_license_file(self, tmp_path: Path) -> None:
        # GIVEN
        (tmp_path / "pyproject.toml").write_text(
            textwrap.dedent(
                """
                [project]
                name = "magic"
                version = "1.0.0"
                license = { file = "LICENSE" }

                [tool.sphinx-theme-builder]
                node-version = "16.13.0"
                """
            )
        )
        (tmp_path / "LICENSE").write_text("MIT")
        (tmp_path / "src" / "magic").mkdir(parents=True)
        (tmp_path / "src" / "magic" / "__init__.py").write_text("")

        # WHEN
        project = Project.from_path(tmp_path)

        # THEN
        expected_licenses = [Path("LICENSE")]
        assert sorted(project.get_license_file_paths()) == expected_licenses
        assert sorted(project.metadata.license_files or []) == expected_licenses

    def test_works_with_license_files(self, tmp_path: Path) -> None:
        # GIVEN
        (tmp_path / "pyproject.toml").write_text(
            textwrap.dedent(
                """
                [project]
                name = "magic"
                version = "1.0.0"
                license-files = ["LICENSE", "licenses/*"]

                [tool.sphinx-theme-builder]
                node-version = "16.13.0"
                """
            )
        )
        (tmp_path / "LICENSE").write_text("MIT")
        (tmp_path / "licenses").mkdir()
        (tmp_path / "licenses" / "LICENSE").write_text("Apache-2.0")
        (tmp_path / "src" / "magic").mkdir(parents=True)
        (tmp_path / "src" / "magic" / "__init__.py").write_text("")

        # WHEN
        project = Project.from_path(tmp_path)

        # THEN
        expected_licenses = [Path("LICENSE"), Path("licenses") / "LICENSE"]
        assert sorted(project.get_license_file_paths()) == expected_licenses
        assert sorted(project.metadata.license_files or []) == expected_licenses

    def test_rejects_import_namespaces(self, tmp_path: Path) -> None:
        # GIVEN
        (tmp_path / "pyproject.toml").write_text(
            textwrap.dedent(
                """
                [project]
                name = "magic"
                version = "1.0.0"
                license = { text = "MIT" }
                import-namespaces = ["something"]

                [tool.sphinx-theme-builder]
                node-version = "16.13.0"
                """
            )
        )
        (tmp_path / "src" / "magic").mkdir(parents=True)
        (tmp_path / "src" / "magic" / "__init__.py").write_text("")

        # WHEN
        with pytest.raises(ImproperProjectMetadata) as ctx:
            Project.from_path(tmp_path)

        # THEN
        error = ctx.value
        assert "unsupported `import-namespaces` key" in error.message
        assert error.code == "invalid-pyproject-toml"

    def test_rejects_mismatched_import_names(self, tmp_path: Path) -> None:
        # GIVEN
        (tmp_path / "pyproject.toml").write_text(
            textwrap.dedent(
                """
                [project]
                name = "magic"
                version = "1.0.0"
                license = { text = "MIT" }
                import-names = ["MAGIC", "notmagic"]

                [tool.sphinx-theme-builder]
                node-version = "16.13.0"
                """
            )
        )
        (tmp_path / "src" / "magic").mkdir(parents=True)
        (tmp_path / "src" / "magic" / "__init__.py").write_text("")
        (tmp_path / "src" / "notmagic").mkdir(parents=True)
        (tmp_path / "src" / "notmagic" / "__init__.py").write_text("")

        # WHEN
        with pytest.raises(ImproperProjectMetadata) as ctx:
            Project.from_path(tmp_path)

        # THEN
        error = ctx.value
        assert "import-names do not match expected value" in error.message
        assert "['magic']" in error.causes[0]
        assert "['MAGIC', 'notmagic']" in error.causes[0]
        assert error.code == "invalid-import-names"

    def test_works_with_matching_import_names(self, tmp_path: Path) -> None:
        # GIVEN
        (tmp_path / "pyproject.toml").write_text(
            textwrap.dedent(
                """
                [project]
                name = "magic-project"
                version = "1.0.0"
                license = { text = "MIT" }
                import-names = ["magic_project"]

                [tool.sphinx-theme-builder]
                node-version = "16.13.0"
                """
            )
        )
        (tmp_path / "src" / "magic_project").mkdir(parents=True)
        (tmp_path / "src" / "magic_project" / "__init__.py").write_text("")

        # WHEN
        project = Project.from_path(tmp_path)

        # THEN
        assert project.snake_name == "magic_project"
        assert project.metadata.import_names == ["magic_project"]



"""Overall workflow tests for `stb new`."""

from click import Group
from click.testing import CliRunner


class TestCLINew:
    def test_no_arguments(self, runner: CliRunner, cli: Group) -> None:
        process = runner.invoke(cli, [])

        assert process.exit_code == 0
        assert process.stdout

    def test_help(self, runner: CliRunner, cli: Group) -> None:
        process = runner.invoke(cli, ["--help"])

        assert process.exit_code == 0
        assert process.stdout

    def test_no_arguments_behaves_same_as_help(
        self, runner: CliRunner, cli: Group
    ) -> None:
        process_one = runner.invoke(cli, [])
        process_two = runner.invoke(cli, ["--help"])

        assert process_one.stdout == process_two.stdout



# Configuration file for the Sphinx documentation builder.
#
# Full list of options can be found in the Sphinx documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

#
# -- Project information -----------------------------------------------------
#

project = "sphinx-theme-builder"
copyright = "2021, Pradyun Gedam"
author = "Pradyun Gedam"

#
# -- General configuration ---------------------------------------------------
#

extensions = [
    # Sphinx's own extensions
    "sphinx.ext.autodoc",
    "sphinx.ext.extlinks",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    # External stuff
    "myst_parser",
    "sphinx_copybutton",
    "sphinx_inline_tabs",
]
templates_path = ["_templates"]

#
# -- Options for extlinks ----------------------------------------------------
#
extlinks = {
    "pypi": ("https://pypi.org/project/%s/", "%s"),
}

#
# -- Options for intersphinx -------------------------------------------------
#
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "sphinx": ("https://www.sphinx-doc.org/en/master", None),
}

#
# -- Options for TODOs -------------------------------------------------------
#
todo_include_todos = True

#
# -- Options for Markdown files ----------------------------------------------
#
myst_enable_extensions = [
    "deflist",
]
myst_heading_anchors = 3

#
# -- Options for HTML output -------------------------------------------------
#

html_theme = "furo"
html_title = "Sphinx Theme Builder"



import contextlib
import errno
import os
import shutil
import stat
import tempfile
from collections.abc import Callable, Generator
from pathlib import Path
from typing import Any

import click
import pytest
from click.testing import CliRunner


# --------------------------------------------------------------------------------------
# Fixtures
@pytest.fixture
def runner() -> CliRunner:
    return CliRunner()


@pytest.fixture
def cli() -> click.Group:
    from sphinx_theme_builder._internal.cli import compose_command_line

    return compose_command_line()


# --------------------------------------------------------------------------------------
# Work around https://github.com/pytest-dev/pytest/issues/7821
@contextlib.contextmanager
def tmpdir() -> Generator[Path, None, None]:
    """Contextmanager to create a temporary directory.  It will be cleaned up
    afterwards.
    """
    tempdir = tempfile.mkdtemp()
    try:
        yield Path(tempdir)
    finally:
        rmtree(tempdir)


def rmtree(path: str) -> None:
    """On windows, rmtree fails for readonly dirs."""

    def handle_remove_readonly(
        func: Callable[..., Any],
        path: str,
        excvalue: BaseException,
    ) -> None:
        if (
            func in (os.rmdir, os.remove, os.unlink)
            and isinstance(excvalue, OSError)
            and excvalue.errno == errno.EACCES
        ):
            for p in (path, os.path.dirname(path)):
                os.chmod(p, os.stat(p).st_mode | stat.S_IWUSR)
            func(path)
        else:
            raise  # noqa

    shutil.rmtree(path, ignore_errors=False, onexc=handle_remove_readonly)


# --------------------------------------------------------------------------------------



"""Overall workflow tests for `stb package`."""

from click import Group
from click.testing import CliRunner


class TestPackageCommand:
    def test_help(self, runner: CliRunner, cli: Group) -> None:
        process = runner.invoke(cli, ["package", "--help"])

        assert process.exit_code == 0, process.stdout
        assert process.stdout



import sys
from unittest import mock

from click import Group
from click.testing import CliRunner


class TestPackageCommand:
    """`stb package`"""

    def test_calls_build(self, runner: CliRunner, cli: Group) -> None:
        with mock.patch("subprocess.run") as mocked_run:
            runner.invoke(cli, ["package"])

        mocked_run.assert_has_calls(
            [mock.call([sys.executable, "-m", "build"], check=True)]
        )



"""Logic for generating distributions from a :ref:`Project`."""

import io
import os
import posixpath
import subprocess
import tarfile
import textwrap
from collections.abc import Callable
from pathlib import Path
from shutil import copyfile

from .errors import STBError
from .nodejs import generate_assets
from .project import Project
from .wheelfile import WheelFile, include_parent_paths


def _get_vcs_tracked_names(path: Path) -> tuple[str, ...] | None:
    if not (path / ".git").is_dir():
        return None

    outb = subprocess.check_output(
        ["git", "ls-files", "--recurse-submodules", "-z"],
        cwd=path,
    )

    tracked_files = [
        os.fsdecode(location) for location in outb.strip(b"\0").split(b"\0") if location
    ]
    return include_parent_paths(tracked_files)


def _sdist_filter(
    project: Project, base: str
) -> Callable[[tarfile.TarInfo], tarfile.TarInfo | None]:
    """Create a filter to pass to tarfile.add, for this project."""
    compiled_assets = project.compiled_assets
    tracked_names = _get_vcs_tracked_names(project.location)
    required_license_paths = {
        os.fsdecode(fp) for fp in project.get_license_file_paths()
    }

    def _filter(tarinfo: tarfile.TarInfo) -> tarfile.TarInfo | None:
        # Include the entry for the root.
        if tarinfo.name == base:
            return tarinfo

        name = tarinfo.name[len(base) + 1 :]

        # Exclude hidden files
        if name.startswith("."):
            return None

        # Exclude compiled pyc files.
        if posixpath.basename(name) == "__pycache__":
            return None

        # Exclude compiled assets.
        if name in compiled_assets:
            return None

        # Never exclude license files.
        if name in required_license_paths:
            return tarinfo

        # Exclude things that are excluded from version control.
        if tracked_names is not None and name not in tracked_names:
            return None

        return tarinfo

    # NOTE: The CLI for build needs to check that the compiled assets we have here, are
    #       not tracked under version control.
    return _filter


#
# External API
#
def generate_source_distribution(
    project: Project,
    *,
    destination: Path,
) -> str:
    """Generate a source distribution for project, and place it inside destination.

    :return: Name of the generated source tarball.
    """
    os.makedirs(destination, exist_ok=True)

    dashed_pair = f"{project.snake_name}-{project.metadata.version}"
    sdist_tarball = destination / f"{dashed_pair}.tar.gz"

    # NOTE: Post Python 3.7 -- can drop the format kwarg, since it'll be the default.
    with tarfile.open(
        name=sdist_tarball, mode="w:gz", format=tarfile.PAX_FORMAT
    ) as tarball:
        # Recursively add the files to this tarball, with an exclusion filter.
        tarball.add(
            project.location,
            arcname=dashed_pair,
            recursive=True,
            filter=_sdist_filter(project, dashed_pair),
        )

        # Write the metadata file.
        metadata_content = project.get_metadata_file_contents().encode()
        tarinfo = tarfile.TarInfo(posixpath.join(dashed_pair, "PKG-INFO"))
        tarinfo.size = len(metadata_content)
        tarball.addfile(tarinfo, io.BytesIO(metadata_content))

    return sdist_tarball.name


def generate_metadata(
    project: Project,
    *,
    destination: Path,
) -> str:
    """Generate the metadata (.dist-info) and place it in `metadata_directory`.

    :return: name of the dist-info directory generated.
    """
    dist_info = (
        destination / f"{project.snake_name}-{project.metadata.version}.dist-info"
    )
    try:
        os.makedirs(dist_info)
    except OSError as error:
        raise STBError(
            message="Metadata directory already exists",
            causes=[],
            hint_stmt=None,
        ) from error

    # Delegated generation.
    (dist_info / "entry_points.txt").write_text(
        project.get_entry_points_contents(), encoding="utf-8"
    )

    # License files go under `licenses/`.
    license_files = project.get_license_file_paths()
    if license_files:
        dest_dir = dist_info / "licenses"
        dest_dir.mkdir()
        for fp in license_files:
            copyfile(fp, dest_dir / fp)

    (dist_info / "METADATA").write_text(
        project.get_metadata_file_contents(), encoding="utf-8"
    )

    # Templated generation.
    (dist_info / "WHEEL").write_text(
        textwrap.dedent(
            """\
            Wheel-Version: 1.0
            Generator: sphinx-theme-builder {version}
            Root-Is-Purelib: true
            Tag: py3-none-any
            """
        ),
        encoding="utf-8",
    )

    return dist_info.name


def generate_wheel_distribution(
    project: Project,
    *,
    destination: Path,
    metadata_directory: Path,
    editable: bool,
) -> str:
    # Generate the JS / CSS assets
    generate_assets(project, production=not editable)

    wheel_path = (
        destination
        / f"{project.snake_name}-{project.metadata.version}-py3-none-any.whl"
    )
    tracked_names = _get_vcs_tracked_names(project.location)

    with WheelFile(
        path=wheel_path,
        compiled_assets=project.compiled_assets,
        tracked_names=tracked_names,
    ) as wheel:
        if editable:
            # Generate a .pth file, for editable installs. There's an enforced src/
            # directory, so this is fairly safe to do.
            wheel.add_string(
                os.fsdecode(project.source_path.resolve()),
                dest=project.snake_name + ".pth",
            )
        else:
            # Add files from the project's src/ directory.
            wheel.add_directory(project.source_path, dest="", base=project.location)

        # Put the metadata at the end.
        # https://www.python.org/dev/peps/pep-0427/#recommended-archiver-features
        wheel.add_directory(metadata_directory, dest=metadata_directory.name, base=None)
        wheel.write_record(dest=metadata_directory.name + "/RECORD")

    return wheel.name



"""A helper class to make it easier to generate a wheel."""

import base64
import hashlib
import os
import posixpath
import zipfile
from dataclasses import dataclass
from pathlib import Path, PurePosixPath
from types import TracebackType
from typing import IO

_HASH_ALGORITHM = "sha256"

# Borrowed from CPython's shutil.py
# https://github.com/python/cpython/blob/v3.9.1/Lib/shutil.py#L52
_WINDOWS = os.name == "nt"
_COPY_BUFSIZE = 1024 * 1024 if _WINDOWS else 64 * 1024


def encode_for_record(hasher: "hashlib._Hash") -> str:
    """Takes a hasher and returns the encoded hash value for the RECORD."""
    return base64.urlsafe_b64encode(hasher.digest()).decode("ascii").rstrip("=")


# Adapted from pradyunsg/installer
# https://github.com/pradyunsg/installer/blob/0.7.0/src/installer/utils.py#L95
def copyfileobj_with_hashing(
    source: IO[bytes],
    dest: IO[bytes],
    hash_algorithm: str,
) -> tuple[str, int]:
    """Copy a buffer while computing the content's hash and size.

    Copies the source buffer into the destination buffer while computing the
    hash of the contents. Adapted from :any:`shutil.copyfileobj`.

    :param source: buffer holding the source data
    :param dest: destination buffer
    :param hash_algorithm: hashing algorithm

    :return: size, hash digest of the contents
    """
    hasher = hashlib.new(hash_algorithm)
    size = 0
    while True:
        buf = source.read(_COPY_BUFSIZE)
        if not buf:
            break
        hasher.update(buf)
        dest.write(buf)
        size += len(buf)

    return encode_for_record(hasher), size


def include_parent_paths(posix_style_paths: list[str]) -> tuple[str, ...]:
    names: set[str] = set()
    for path_str in posix_style_paths:
        path = PurePosixPath(path_str)
        names.update(parent.as_posix() for parent in path.parents)
        names.add(path_str)
    return tuple(names)


@dataclass
class RecordEntry:
    """A single entry in a RECORD file."""

    path: str
    hash_value: str
    hash_algorithm: str
    size: str

    def to_line(self) -> str:
        if self.hash_value:
            return ",".join(
                (self.path, f"{self.hash_algorithm}={self.hash_value}", self.size)
            )
        else:
            assert not self.hash_algorithm
            return ",".join((self.path, "", self.size))


class WheelFile:
    """A helper class to generate wheels."""

    def __init__(
        self,
        *,
        path: Path,
        tracked_names: tuple[str, ...] | None,
        compiled_assets: tuple[str, ...],
    ) -> None:
        self._path = path
        self._tracked_names = tracked_names
        self._compiled_assets = compiled_assets
        self._compiled_assets_and_parents = include_parent_paths(list(compiled_assets))
        self._zipfile = zipfile.ZipFile(path, mode="w")
        self._records: list[RecordEntry] = []

    def __enter__(self) -> "WheelFile":
        return self

    def __exit__(
        self,
        type_: type[BaseException] | None,
        value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        self._zipfile.close()

    def _exclude(self, path: Path, *, base: Path | None) -> bool:
        # Exclude compiled pyc files.
        if path.name == "__pycache__":
            return True

        # Include all not-based-on-source-tree files.
        if base is None:
            return False

        normalised_path = path.relative_to(base).as_posix()

        # Definitely include compiled assets.
        if normalised_path in self._compiled_assets_and_parents:
            return False

        for asset_path in self._compiled_assets:
            if normalised_path.startswith(asset_path + "/"):
                return False

        # Exclude things that are excluded from version control.
        if self._tracked_names is not None:
            if normalised_path not in self._tracked_names:
                return True

        # If we're here, we've excluded all the files that needed to be excluded.
        return False

    @property
    def name(self) -> str:
        """Name of the wheel."""
        return self._path.name

    def write_record(self, *, dest: str) -> None:
        """Write the record, in the provided destination.

        :param dest: The exact ``{package}-{version}.dist-info/RECORD`` to write to.
        """
        assert self._zipfile.fp is not None

        self._records.append(RecordEntry(dest, "", "", ""))
        lines = [record.to_line() for record in self._records]
        self._zipfile.writestr(dest, data="\n".join(lines))

    def add_string(self, content: str, *, dest: str) -> None:
        """Add a file at ``dest``, with the given ``content``."""
        assert self._zipfile.fp is not None

        data = content.encode()
        hasher = hashlib.new(_HASH_ALGORITHM, data=data)

        self._zipfile.writestr(dest, data=data)
        self._records.append(
            RecordEntry(
                path=dest,
                hash_algorithm=_HASH_ALGORITHM,
                hash_value=encode_for_record(hasher),
                size=str(len(data)),
            )
        )

    def add_file(self, file: Path, *, dest: str, base: Path | None) -> None:
        """Add a file at ``dest``, with the contents of ``file``."""
        assert self._zipfile.fp is not None

        if self._exclude(file, base=base):
            return

        # Copy the file object.
        zipinfo = zipfile.ZipInfo.from_file(file, dest)
        with file.open("rb") as source_stream:
            with self._zipfile.open(zipinfo, "w") as dest_stream:
                hash_value, size = copyfileobj_with_hashing(
                    source_stream, dest_stream, hash_algorithm=_HASH_ALGORITHM
                )
        self._records.append(
            RecordEntry(
                path=dest,
                hash_algorithm=_HASH_ALGORITHM,
                hash_value=hash_value,
                size=str(size),
            )
        )

    def add_directory(self, directory: Path, *, dest: str, base: Path | None) -> None:
        """Add the directory to the archive, recursively."""
        assert directory.is_dir()
        assert self._zipfile.fp is not None

        if self._exclude(directory, base=base):
            return

        for item in sorted(directory.iterdir()):
            if item.is_dir():
                self.add_directory(
                    item, dest=posixpath.join(dest, item.name), base=base
                )
            else:
                self.add_file(item, dest=posixpath.join(dest, item.name), base=base)



"""Development automation."""

import glob
import os
import shutil
import tempfile

import nox

PACKAGE_NAME = "sphinx_theme_builder"
nox.options.sessions = ["lint", "docs"]


#
# Development Sessions
#
@nox.session(name="docs-live", reuse_venv=True)
def docs_live(session):
    session.install("-e", ".[cli]")
    session.install("-r", "docs/requirements.txt")
    session.install("sphinx-autobuild")

    with tempfile.TemporaryDirectory() as destination:
        session.run(
            "sphinx-autobuild",
            # for sphinx-autobuild
            "--watch=README.md",
            "--port=0",
            "--open-browser",
            # for sphinx
            "-b=dirhtml",
            "-a",
            "docs/",
            destination,
        )


@nox.session(reuse_venv=True)
def docs(session):
    session.install("-e", ".[cli]")
    session.install("-r", "docs/requirements.txt")

    # Generate documentation into `build/docs`
    session.run(
        "sphinx-build",
        "-b",
        "dirhtml",
        "-W",
        "--keep-going",
        "-v",
        "docs/",
        "build/docs",
    )


@nox.session(reuse_venv=True)
def lint(session: nox.Session):
    session.install("prek", "mypy")

    args = list(session.posargs)
    args.append("--all-files")
    if "CI" in os.environ:
        args.append("--show-diff-on-failure")

    session.run("prek", "run", *args)
    session.notify("mypy")


@nox.session
def mypy(session: nox.Session):
    session.install(".[cli]", "mypy", "-r", "tests/requirements.txt")
    session.run("mypy", "src", "tests", "--strict")


@nox.session(
    python=["3.10", "3.11", "3.12", "3.13", "3.14"],
    reuse_venv=True,
)
def test(session):
    if os.environ.get("CI"):
        session.install(".[cli]")
    else:
        session.install("-e", ".[cli]")

    session.install("-r", "tests/requirements.txt")

    args = session.posargs or (
        "--cov=sphinx_theme_builder",
        "--cov-branch",
        "--cov-report=term-missing",
        "--cov-report=html",
    )

    session.run(
        "pytest",
        "--pspec",
        *args,
    )


def get_next_version(release_version: str) -> str:
    major, minor, _ = release_version.split(".", 2)
    next_minor = str(int(minor) + 1)
    return ".".join([major, next_minor, "0", "dev1"])


@nox.session
def release(session):
    [release_version] = session.posargs

    version_file = f"src/{PACKAGE_NAME}/__init__.py"
    allowed_upstreams = [
        f"git@github.com:pradyunsg/{PACKAGE_NAME.replace('_', '-')}.git"
    ]

    next_version = get_next_version(release_version)

    session.install("build", "twine", "release-helper")

    # Sanity Checks
    session.run("release-helper", "version-check-validity", release_version)
    session.run("release-helper", "version-check-validity", next_version)
    session.run("release-helper", "directory-check-empty", "dist")

    session.run("release-helper", "git-check-branch", "main")
    session.run("release-helper", "git-check-clean")
    session.run("release-helper", "git-check-tag", release_version, "--does-not-exist")
    session.run("release-helper", "git-check-remote", "origin", *allowed_upstreams)

    # Prepare release commit
    session.run("release-helper", "version-bump", version_file, release_version)
    session.run("git", "add", version_file, external=True)

    session.run(
        "git", "commit", "-m", f"Prepare release: {release_version}", external=True
    )

    # Build the package
    if os.path.exists("build"):
        shutil.rmtree("build")
    session.run("python", "-m", "build")
    session.run("twine", "check", *glob.glob("dist/*"))

    # Tag the commit
    session.run(
        "git",
        "tag",
        release_version,
        "-m",
        f"Release {release_version}",
        "-s",
        external=True,
    )

    # Prepare back-to-development commit
    session.run("release-helper", "version-bump", version_file, next_version)
    session.run("git", "add", version_file, external=True)
    session.run("git", "commit", "-m", "Back to development", external=True)

    # Push the commits and tag.
    session.run("git", "push", "origin", "main", release_version, external=True)

    # Upload the distributions.
    session.run("twine", "upload", *glob.glob("dist/*"))



from unittest import mock

from click import Group
from click.testing import CliRunner


class TestCompileCommand:
    """`stb compile`"""

    def test_calls_generate_assets(self, runner: CliRunner, cli: Group) -> None:
        with (
            mock.patch(
                "sphinx_theme_builder._internal.cli.compile.generate_assets"
            ) as mocked_generate_assets,
            mock.patch(
                "sphinx_theme_builder._internal.cli.compile.Project"
            ) as mocked_project,
        ):
            with runner.isolated_filesystem():
                process = runner.invoke(cli, ["compile"])

        assert process.exit_code == 0, process

        mocked_generate_assets.assert_has_calls(
            [
                mock.call(mocked_project.from_cwd(), production=False),
            ]
        )

    def test_calls_generate_assets_in_production(
        self, runner: CliRunner, cli: Group
    ) -> None:
        with (
            mock.patch(
                "sphinx_theme_builder._internal.cli.compile.generate_assets"
            ) as mocked_generate_assets,
            mock.patch(
                "sphinx_theme_builder._internal.cli.compile.Project"
            ) as mocked_project,
        ):
            with runner.isolated_filesystem():
                process = runner.invoke(cli, ["compile", "--production"])

        assert process.exit_code == 0, process

        mocked_generate_assets.assert_has_calls(
            [
                mock.call(mocked_project.from_cwd(), production=True),
            ]
        )
